{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE ViewPatterns #-}

import Test.QuickCheck
import Test.QuickCheck.Function
{-
Wir definieren eine Typklasse Product für Produkte von zwei Typen a und b.
Dann machen wir gewöhnliche Haskell-Paare zu einer Instanz dieser Klasse.
-}
class Product prod where 
  proj1 :: (prod a b) -> a
  proj2 :: (prod a b) -> b
  factorize :: (c -> a) -> (c -> b) -> (c -> (prod a b))
  --Must satisfy:
  -- proj1 . factorize f g = f
  -- proj2 . factorize f g = g
  -- For all f :: (c -> a ) -> (c -> b) -> (c -> (prod a b)) we have:
  -- f = factorize

{-
Das gewöhnliche Haskell Tupel ist eine Instanz von Product
-}
instance Product (,) where
  proj1 :: (a,b) -> a
  proj1 = fst
  proj2 :: (a,b) -> b
  proj2 = snd
  factorize :: (c -> a) -> (c -> b) -> (c -> (a,b))
  factorize f g = \x -> (f x, g x)

{-
Wir führen einen neuen Datentyp "Pair" ein.
-}
data Pair a b = MkPair b a

{-
Machen Sie Pair zu einer Instanz von Product.
Danach können Sie die Tests auskommentieren und ihre Instanz testen.
Cp. http://austinrochford.com/posts/2014-05-27-quickcheck-laws.html
-}

instance Product Pair where
  proj1 = undefined
  proj2 = undefined
  factorize = undefined

--Tests
prop_firstLaw :: (Eq a) => Fun c a -> Fun c b -> c -> Bool
prop_firstLaw (apply -> f) (apply -> g) x = (((proj1 :: Pair a b -> a ). (factorize f g))  x)  == (f x)

prop_secondLaw :: (Eq b) => Fun c a -> Fun c b -> c -> Bool
prop_secondLaw (apply -> f) (apply -> g) x = (((proj2 :: Pair a b -> b ). (factorize f g))  x)  == (g x)


{-
Die Typklasse ProdOfInteger spezialisiert die obige Klasse für den Typ von Paaren von Integers.
Wir können gewöhnliche Haskell-Paare von zwei Integers, i.e. (Integer, Integer), zu
einer Instanz machen.
-}

class ProdOfInteger poi where
  proj1' :: poi -> Integer
  proj2' :: poi -> Integer
  factorize' :: (c -> Integer) -> (c -> Integer) -> (c -> poi)
  --Must satisfy:
  -- proj1' . factorize' f g = f
  -- proj2' . factorize' f g = g
  -- For all f :: (c -> a ) -> (c -> b) -> (c -> poi) we have:
  -- f = factorize'
  
instance ProdOfInteger (Integer,Integer) where
  proj1' :: (Integer,Integer) -> Integer
  proj1' = fst
  proj2' :: (Integer,Integer) -> Integer
  proj2' = snd
  factorize' :: (c -> Integer) -> (c -> Integer) -> (c -> (Integer, Integer))
  factorize' f g = \x -> (f x, g x)
  
{-
Hilfsfunktionen
-}

type Nat = Integer

natToInt :: Nat -> Integer
natToInt n = if even n then n `div` 2 else -((n + 1) `div` 2)

intToNat :: Integer -> Nat
intToNat z = if (z >= 0) then 2 * z else ((-2 * z) - 1) 

cantorpair :: Nat -> Nat -> Nat
cantorpair m n  = (((m + n) * (m + n + 1)) `div` 2)+ n

cantorFst :: Nat -> Nat
cantorFst p = j - (p - ((j * (j + 1)) `div` 2))
    where j = floor ( (sqrt $ 2 * (fromIntegral p) + 0.25) - 0.5)
          
cantorSnd :: Nat -> Nat
cantorSnd p = p - ((j * (j + 1)) `div` 2)
    where j = floor ( (sqrt $ 2 * (fromIntegral p) + 0.25) - 0.5)

{-
Die folgenden beiden Funktionen bilden einen Isomorphismus zwischen Integer und (Integer, Integer)
-}
i2ii :: Integer -> (Integer, Integer)
i2ii x = (natToInt . cantorFst . intToNat $ x, natToInt . cantorSnd . intToNat $ x)

ii2i :: (Integer, Integer) -> Integer
ii2i (x,y) = natToInt $ cantorpair (intToNat x) (intToNat y)

{-
Benutzen Sie die beiden Funktionen i2ii und ii2i um Integer zu einer Instanz von ProdOfInteger zu machen.
-}

instance ProdOfInteger Integer where
  proj1' = undefined
  proj2' = undefined
  factorize' f g x =  undefined


--Tests
prop_firstLaw' :: Fun c Integer -> Fun c Integer -> c -> Bool
prop_firstLaw' (apply -> f) (apply -> g) x = (((proj1' :: Integer -> Integer ). (factorize' f g))  x)  == (f x)

prop_secondLaw' :: Fun c Integer -> Fun c Integer -> c -> Bool
prop_secondLaw' (apply -> f) (apply -> g) x = (((proj2' :: Integer -> Integer ). (factorize' f g))  x)  == (g x)



main :: IO ()
main = do
  print "Tests:"

  --Tests
  quickCheck (prop_firstLaw  :: Fun Integer Integer -> Fun Integer Integer -> Integer -> Bool)
  quickCheck (prop_secondLaw :: Fun Integer Integer -> Fun Integer Integer -> Integer -> Bool)

  --Tests
  --quickCheck (prop_firstLaw'  :: Fun Integer Integer -> Fun Integer Integer -> Integer -> Bool)
  --quickCheck (prop_secondLaw' :: Fun Integer Integer -> Fun Integer Integer -> Integer -> Bool)
